@@ -1,2 +1,2 @@
-from .client import Client
+from .client import Client, CopyOptions
 from .errors import PulpException, TaskFailedException
