@@ -17,12 +17,14 @@ from ..criteria import Criteria
 from ..model import (
     Repository,
     FileRepository,
+    YumRepository,
     MaintenanceReport,
     Distributor,
     Unit,
     Task,
 )
 from ..log import TimedLogger
+from ..util import dict_put
 from .search import search_for_criteria
 from .errors import PulpException
 from .poller import TaskPoller
@@ -30,6 +32,7 @@ from . import retry
 from .humanize_compat import naturalsize
 
 from .ud_mappings import compile_ud_mappings
+from .copy import CopyOptions
 
 
 LOG = logging.getLogger("pubtools.pulplib")
@@ -138,12 +141,23 @@ class Client(object):
                 If more than this number of tasks are running, the client will wait before triggering more.
                 This can be used to ensure no single client overwhelms the Pulp server.
 
+            int threads
+                A hint for the number of threads the client should use in order to
+                make concurrent requests to Pulp.
+
+                This number should be considered as more of a scaling factor than
+                a precise thread count. The client in practice will use more than
+                this number of threads.
+
             object auth, cert, headers, max_redirects, params, proxies, verify
                 Any of these arguments, if provided, are used to initialize
                 :class:`requests.Session` objects used by the client.
 
                 These may be used, for example, to configure the credentials
                 for the Pulp server or to use an alternative CA bundle.
+
+        .. versionadded:: 2.31.0
+            Added the ``threads`` argument.
         """
         self._url = url
 
@@ -163,7 +177,10 @@ class Client(object):
             if arg in kwargs:
                 self._session_kwargs[arg] = kwargs.pop(arg)
 
-        _task_throttle = kwargs.pop("task_throttle", self._TASK_THROTTLE)
+        task_throttle = kwargs.pop("task_throttle", self._TASK_THROTTLE)
+        threads = kwargs.pop("threads", None)
+        if threads is not None and threads < 1:
+            threads = 1
 
         if kwargs:
             raise TypeError(
@@ -181,7 +198,8 @@ class Client(object):
         #   it's not successful
         self._request_executor = (
             Executors.thread_pool(
-                name="pubtools-pulplib-requests", max_workers=self._REQUEST_THREADS
+                name="pubtools-pulplib-requests",
+                max_workers=threads or self._REQUEST_THREADS,
             )
             .with_map(self._unpack_response)
             .with_retry(retry_policy=self._RETRY_POLICY)
@@ -193,7 +211,7 @@ class Client(object):
         # - HTTP requests don't happen here and are still submitted via
         #   request_executor (hence no retry needed on this executor)
         self._upload_executor = Executors.thread_pool(
-            name="pubtools-pulplib-uploads", max_workers=self._UPLOAD_THREADS
+            name="pubtools-pulplib-uploads", max_workers=threads or self._UPLOAD_THREADS
         )
 
         # executor for issuing HTTP requests to Pulp which spawn tasks:
@@ -206,12 +224,13 @@ class Client(object):
         poller = TaskPoller(self._new_session(), self._url)
         self._task_executor = (
             Executors.thread_pool(
-                name="pubtools-pulplib-tasks", max_workers=self._REQUEST_THREADS
+                name="pubtools-pulplib-tasks",
+                max_workers=threads or self._REQUEST_THREADS,
             )
             .with_map(self._unpack_response)
             .with_map(self._log_spawned_tasks)
             .with_poll(poller, cancel_fn=poller.cancel)
-            .with_throttle(_task_throttle)
+            .with_throttle(task_throttle)
             .with_retry(retry_policy=self._RETRY_POLICY)
         )
 
@@ -300,7 +319,9 @@ class Client(object):
             )
         )
 
-    def copy_content(self, from_repository, to_repository, criteria=None):
+    def copy_content(
+        self, from_repository, to_repository, criteria=None, options=CopyOptions()
+    ):
         """Copy content from one repository to another.
 
         Args:
@@ -314,6 +335,13 @@ class Client(object):
                 A criteria object used to find units for copy.
                 If None, all units in the source repo will be copied.
 
+            options (:class:`~pubtools.pulplib.CopyOptions`)
+                Options influencing the copy.
+
+                Some options may be specific to certain repository and content
+                types. Options which are not applicable to this copy will be
+                ignored.
+
         Returns:
             Future[list[:class:`~pubtools.pulplib.Task`]]
                 A future which is resolved when the copy completes.
@@ -323,9 +351,27 @@ class Client(object):
                 typically will have only a subset of available fields.
 
         .. versionadded:: 2.17.0
+
+        .. versionadded:: 2.29.0
+            Added the ``options`` argument.
         """
+
+        raw_options = {}
+
+        if (
+            isinstance(to_repository, YumRepository)
+            and options.require_signed_rpms is not None
+        ):
+            dict_put(
+                raw_options,
+                "override_config.require_signature",
+                options.require_signed_rpms,
+            )
+
         return f_proxy(
-            self._do_associate(from_repository.id, to_repository.id, criteria)
+            self._do_associate(
+                from_repository.id, to_repository.id, criteria, raw_options
+            )
         )
 
     def update_content(self, unit):
@@ -369,6 +415,39 @@ class Client(object):
 
         return out
 
+    def update_repository(self, repository):
+        """Update mutable fields on an existing repository.
+
+        Args:
+            repository (:class:`~pubtools.pulplib.Repository`)
+                A repository to be updated.
+
+                Only those fields documented as *mutable* will have any effect during
+                the update (e.g. ``product_versions``). Other fields cannot be updated
+                using this library.
+
+        Returns:
+            Future
+                A future which is resolved with a value of ``None`` once the
+                repository has been updated.
+
+        .. versionadded:: 2.29.0
+        """
+
+        url = os.path.join(self._url, "pulp/api/v2/repositories/%s/" % repository.id)
+
+        update = {"delta": {"notes": repository._mutable_notes}}
+        out = self._request_executor.submit(
+            self._do_request, method="PUT", url=url, json=update
+        )
+
+        # The Pulp API may actually return an updated version of the repository,
+        # but for consistency with update_content we won't return it.
+        # The caller can re-fetch if desired.
+        out = f_map(out, lambda _: None)
+
+        return out
+
     def _search_content_with_server_type_ids(self, criteria, server_type_ids):
         prepared_search = search_for_criteria(criteria, Unit, None)
         type_ids = prepared_search.type_ids
@@ -454,16 +533,27 @@ class Client(object):
                     "filters": prepared_search.filters,
                 }
             }
+
+            if prepared_search.unit_fields is not None:
+                search["criteria"]["fields"] = sorted(
+                    set(
+                        [field.pulp_field_name for field in prepared_search.unit_fields]
+                    )
+                )
+
             search.update(search_options or {})
 
             if search_type == "search/units":
                 # Unit searches need a little special handling:
                 # - serialization might have extracted some type_ids
-                # - filters should be wrapped under 'unit'
+                # - filters/fields should be wrapped under 'unit'
                 #   (we do not support searching on associations right now)
                 if prepared_search.type_ids:
                     search["criteria"]["type_ids"] = prepared_search.type_ids
-                search["criteria"]["filters"] = {"unit": search["criteria"]["filters"]}
+
+                for elem in ("filters", "fields"):
+                    if elem in search["criteria"]:
+                        search["criteria"][elem] = {"unit": search["criteria"][elem]}
 
             searches.append(search)
             responses.append(self._do_search(url, search))
@@ -662,12 +752,14 @@ class Client(object):
             self._do_request, method="POST", url=url, json=body
         )
 
-    def _do_associate(self, src_repo_id, dest_repo_id, criteria=None):
+    def _do_associate(self, src_repo_id, dest_repo_id, criteria=None, raw_options=None):
         url = os.path.join(
             self._url, "pulp/api/v2/repositories/%s/actions/associate/" % dest_repo_id
         )
 
-        pulp_search = search_for_criteria(criteria, type_hint=Unit, type_ids_accum=None)
+        pulp_search = search_for_criteria(
+            criteria, type_hint=Unit, unit_type_accum=None
+        )
 
         body = {"source_repo_id": src_repo_id, "criteria": {}}
         if pulp_search.type_ids:
@@ -675,6 +767,8 @@ class Client(object):
         if pulp_search.filters:
             body["criteria"]["filters"] = {"unit": pulp_search.filters}
 
+        body.update(raw_options or {})
+
         LOG.debug("Submitting %s associate: %s", url, body)
 
         return self._task_executor.submit(
@@ -689,10 +783,23 @@ class Client(object):
         # use type hint=Unit so that if type_ids are the goal here
         # then we will get back a properly prepared PulpSearch with
         # a populated type_ids field
-        pulp_search = search_for_criteria(criteria, type_hint=Unit, type_ids_accum=None)
+        pulp_search = search_for_criteria(
+            criteria, type_hint=Unit, unit_type_accum=None
+        )
 
         body = {"criteria": {"type_ids": pulp_search.type_ids}}
 
+        if pulp_search.filters:
+            # Filters to remove content are only effective when provided
+            # with a type_id. Otherwise filters are skipped and all the
+            # content is removed.
+            if not pulp_search.type_ids:
+                raise ValueError(
+                    "Criteria to remove_content must specify at least one unit type!"
+                )
+            else:
+                body["criteria"]["filters"] = {"unit": pulp_search.filters}
+
         LOG.debug("Submitting %s unassociate: %s", url, body)
 
         return self._task_executor.submit(
