@@ -14,9 +14,17 @@ except AttributeError:  # pragma: no cover
     # python 2
     Iterable = collections.Iterable  # pylint: disable=invalid-name
 
+from frozenlist2 import frozenlist
+
 from pubtools.pulplib._impl import compat_attr as attr
 
 from .model.unit import type_ids_for_class
+from .model.attr import PULP2_FIELD
+
+
+FieldNamePair = collections.namedtuple(
+    "FieldNamePair", ["model_field_name", "pulp_field_name"]
+)
 
 
 class Criteria(object):
@@ -102,25 +110,73 @@ class Criteria(object):
         return FieldMatchCriteria(field_name, field_value)
 
     @classmethod
-    def with_unit_type(cls, unit_type):
+    def with_unit_type(cls, unit_type, **kwargs):
         """Args:
             unit_type (class)
                 A subclass of :class:`~pubtools.pulplib.Unit`.
 
+            unit_fields (Iterable[str])
+                Names of the desired field(s) to include in response to a search
+                using this criteria. If omitted, all fields are included.
+
+                Some fields will always be included even when not requested.
+
         Returns:
             Criteria
-                criteria for finding units of type ``unit_type``.
+                criteria for finding units of type ``unit_type`` populated with
+                (at least) the fields from ``unit_fields``.
 
         .. versionadded:: 2.14.0
+
+        .. versionadded:: 2.33.0
+            Introduced ``unit_fields``.
         """
 
-        # This is just a thin wrapper for searching on content_type_id which allows
+        # This is mainly a thin wrapper for searching on content_type_id which allows
         # the caller to avoid having to handle the (unit class <=> type id) mapping.
         type_ids = type_ids_for_class(unit_type)
         if not type_ids:
             raise TypeError("Expected a Unit type, got: %s" % repr(unit_type))
 
-        return FieldMatchCriteria("content_type_id", Matcher.in_(type_ids))
+        unit_fields = kwargs.pop("unit_fields", None)
+        if unit_fields is not None:
+            # We have some non-default set of fields to query.
+            # We must do the following:
+            # - ensure that we also include all the mandatory fields for this model
+            # - generate pairs of (model field, pulp field) names, as we'll need both
+
+            model_field_names = set(unit_fields)
+
+            model_fields_dict = attr.fields_dict(unit_type)
+            for field in model_fields_dict.values():
+                if field.default is attr.NOTHING:
+                    # No default => it's mandatory to use this field
+                    model_field_names.add(field.name)
+
+            # Now build up (model, pulp) pairs
+            pairs = set()
+            for model_field_name in model_field_names:
+                # The default/fallback is to assume that the pulp field is named the
+                # same as the model field...
+                pulp_field_name = model_field_name
+
+                # ... but if the field exists and declares an explicit PULP2_FIELD then we
+                # use that instead
+                field = model_fields_dict.get(model_field_name)
+                if field and field.metadata.get(PULP2_FIELD):
+                    # We have a defined pulp field, so use it.
+                    # Note we only care about the first component of the field, because that's
+                    # all the granularity supported by Pulp search API (e.g. if actual field
+                    # is pulp_user_metadata.description, we need to query pulp_user_metadata).
+                    pulp_field_name = field.metadata[PULP2_FIELD].split(".")[0]
+
+                pairs.add(FieldNamePair(model_field_name, pulp_field_name))
+
+            unit_fields = tuple(sorted(pairs))
+
+        return UnitTypeMatchCriteria(
+            "content_type_id", Matcher.in_(type_ids), unit_fields
+        )
 
     @classmethod
     def with_field_in(cls, field_name, field_value):
@@ -285,7 +341,7 @@ class Matcher(object):
         return self
 
 
-@attr.s
+@attr.s(frozen=True)
 class RegexMatcher(Matcher):
     _pattern = attr.ib()
 
@@ -307,7 +363,7 @@ class RegexMatcher(Matcher):
         return "=~/%s/" % self._pattern
 
 
-@attr.s
+@attr.s(frozen=True)
 class EqMatcher(Matcher):
     _value = attr.ib()
 
@@ -318,15 +374,15 @@ class EqMatcher(Matcher):
         return "==%s" % repr(self._value)
 
 
-@attr.s
-class InMatcher(Matcher):
-    _values = attr.ib()
+def iterable_nonstr_then_frozenlist(values):
+    if isinstance(values, Iterable) and not isinstance(values, six.string_types):
+        return frozenlist(values)
+    raise ValueError("Must be an iterable: %s" % repr(values))
 
-    @_values.validator
-    def _check_values(self, _, values):
-        if isinstance(values, Iterable) and not isinstance(values, six.string_types):
-            return
-        raise ValueError("Must be an iterable: %s" % repr(values))
+
+@attr.s(frozen=True)
+class InMatcher(Matcher):
+    _values = attr.ib(converter=iterable_nonstr_then_frozenlist)
 
     def _map(self, fn):
         return attr.evolve(self, values=[fn(x) for x in self._values])
@@ -335,13 +391,13 @@ class InMatcher(Matcher):
         return " IN %s" % repr(self._values)
 
 
-@attr.s
+@attr.s(frozen=True)
 class ExistsMatcher(Matcher):
     def __str__(self):
         return " EXISTS"
 
 
-@attr.s
+@attr.s(frozen=True)
 class LessThanMatcher(Matcher):
     _value = attr.ib()
 
@@ -367,7 +423,7 @@ def coerce_to_matcher(value):
     return EqMatcher(value)
 
 
-@attr.s
+@attr.s(frozen=True)
 class FieldMatchCriteria(Criteria):
     _field = attr.ib()
     _matcher = attr.ib(converter=coerce_to_matcher)
@@ -381,7 +437,14 @@ class FieldMatchCriteria(Criteria):
         return out
 
 
-@attr.s
+@attr.s(frozen=True)
+class UnitTypeMatchCriteria(FieldMatchCriteria):
+    # This specialization of FieldMatchCriteria is used to match on unit types
+    # while also keeping info on the fields of interest to the user.
+    _unit_fields = attr.ib()
+
+
+@attr.s(frozen=True)
 class AndCriteria(Criteria):
     _operands = attr.ib()
 
@@ -395,7 +458,7 @@ class AndCriteria(Criteria):
         return "(" + " AND ".join([str(o) for o in self._operands]) + ")"
 
 
-@attr.s
+@attr.s(frozen=True)
 class OrCriteria(Criteria):
     _operands = attr.ib()
 
