@@ -1,11 +1,14 @@
 import datetime
 import logging
+import warnings
+import json
+from functools import partial
 
 from attr import validators, asdict
 from frozenlist2 import frozenlist
 from more_executors.futures import f_proxy, f_map, f_flat_map
 
-from ..attr import pulp_attrib
+from ..attr import pulp_attrib, PULP2_FIELD, PULP2_MUTABLE
 from ..common import PulpObject, Deletable, DetachedException
 from ..convert import frozenlist_or_none_converter
 from ..distributor import Distributor
@@ -13,6 +16,7 @@ from ...criteria import Criteria, Matcher
 from ...schema import load_schema
 from ... import compat_attr as attr
 from ...hooks import pm
+from ...util import dict_put, lookup, ABSENT
 
 
 LOG = logging.getLogger("pubtools.pulplib")
@@ -131,6 +135,27 @@ class SyncOptions(object):
     """
 
 
+def pv_converter(versions):
+    # Converter for values in a product_versions field.
+    #
+    # We try to sort numerically while decomposing dotted versions into
+    # their components, so e.g. "8.10" is sorted later than "8.8".
+    # However, we don't know for sure which version strings might be stored
+    # in the field (or even non-strings), so we can fall back to a generic
+    # string sort.
+
+    # Everything is initially interpreted as a string regardless of how
+    # it was stored.
+    versions = [str(v) for v in versions]
+
+    try:
+        return sorted(
+            versions, key=lambda version: [int(c) for c in str(version).split(".")]
+        )
+    except ValueError:
+        return sorted(versions)
+
+
 @attr.s(kw_only=True, frozen=True)
 class Repository(PulpObject, Deletable):
     """Represents a Pulp repository."""
@@ -237,6 +262,75 @@ class Repository(PulpObject, Deletable):
     content_set = pulp_attrib(default=None, type=str, pulp_field="notes.content_set")
     """Name of content set that is associated with this repository."""
 
+    arch = pulp_attrib(default=None, type=str, pulp_field="notes.arch")
+    """The primary architecture of content within this repository (e.g. 'x86_64').
+
+    .. versionadded:: 2.29.0
+    """
+
+    platform_full_version = pulp_attrib(
+        default=None, type=str, pulp_field="notes.platform_full_version"
+    )
+    """A version string associated with the repository.
+
+    This field should be used with care, as the semantics are not well defined.
+    It is often, but not always, equal to the $releasever yum variable associated
+    with a repository.
+
+    Due to the unclear meaning of this field, it's strongly recommended to avoid
+    making use of it in any new code.
+
+    .. versionadded:: 2.29.0
+    """
+
+    product_versions = pulp_attrib(
+        default=None,
+        type=list,
+        pulp_field="notes.product_versions",
+        pulp_py_converter=json.loads,
+        py_pulp_converter=partial(json.dumps, separators=(",", ":")),
+        converter=partial(frozenlist_or_none_converter, map_fn=pv_converter),
+        mutable=True,
+    )
+    """A list of product versions associated with this repository.
+
+    The versions found in this list are derived from the product versions found in any
+    product certificates (productid) historically uploaded to this repository and
+    related repositories.
+
+    This field is **mutable** and may be set by :meth:`~Client.update_repository`.
+
+    .. versionadded:: 2.29.0
+    """
+
+    include_in_download_service = pulp_attrib(
+        default=False,
+        type=bool,
+        mutable=True,
+        pulp_field="notes.include_in_download_service",
+        pulp_py_converter=lambda x: x == "True",
+        py_pulp_converter=str,
+    )
+    """Flag indicating whether the repository is visible in production instance
+    of download service.
+
+    .. versionadded:: 2.34.0
+    """
+
+    include_in_download_service_preview = pulp_attrib(
+        default=False,
+        type=bool,
+        mutable=True,
+        pulp_field="notes.include_in_download_service_preview",
+        pulp_py_converter=lambda x: x == "True",
+        py_pulp_converter=str,
+    )
+    """Flag indicating whether the repository is visible in staging instance
+    of download service.
+
+    .. versionadded:: 2.34.0
+    """
+
     @distributors.validator
     def _check_repo_id(self, _, value):
         # checks if distributor's repository id is same as the repository it
@@ -258,6 +352,38 @@ class Repository(PulpObject, Deletable):
             out[dist.id] = dist
         return out
 
+    @classmethod
+    def _mutable_note_fields(cls):
+        # Returns the subset of fields on this class which are stored under the
+        # notes dict and considered mutable, and thus can potentially be updated.
+        return [
+            fld
+            for fld in attr.fields(cls)
+            if fld.metadata.get(PULP2_FIELD, "").startswith("notes.")
+            and fld.metadata.get(PULP2_MUTABLE)
+        ]
+
+    @property
+    def _mutable_notes(self):
+        # Returns notes dict containing only mutable notes, appropriate
+        # for updating the repo.
+
+        # Get self in raw Pulp form.
+        self_raw = self._to_data()
+
+        # Make a filtered view keeping only mutable note fields.
+        out = {}
+
+        for field in self._mutable_note_fields():
+            pulp_field = field.metadata.get(PULP2_FIELD)
+            pulp_value = lookup(self_raw, pulp_field)
+
+            if pulp_value is not ABSENT:
+                dict_put(out, pulp_field, pulp_value)
+
+        # Return only the notes portion.
+        return out.get("notes") or {}
+
     def distributor(self, distributor_id):
         """Look up a distributor by ID.
 
@@ -469,15 +595,17 @@ class Repository(PulpObject, Deletable):
             )
         )
 
-    def remove_content(self, **kwargs):
+    def remove_content(self, criteria=None, **kwargs):
         """Remove all content of requested types from this repository.
 
         Args:
-            type_ids (list[str])
-                IDs of content type(s) to be removed.
-                See :meth:`~pubtools.pulplib.Client.get_content_type_ids`.
+            criteria (:class:`~pubtools.pulplib.Criteria`)
+                A criteria object used to filter the contents for removal.
 
-                If omitted, content of all types will be removed.
+                Type IDs must be included in the criteria with any other filters.
+                If omitted, filter criteria will be ignored and all the content will
+                be removed.
+                If criteria is omitted, all the content will be removed.
 
         Returns:
             Future[list[:class:`~pubtools.pulplib.Task`]]
@@ -498,33 +626,30 @@ class Repository(PulpObject, Deletable):
         if not self._client:
             raise DetachedException()
 
-        # Note: we use dynamic kwargs because it's very likely that a future
-        # version of this method will support some "criteria".  Let's not fix the
-        # argument order at least until then.
+        # Type IDs are must for the criteria filter to be effective. It must be included
+        # in the criteria. Type IDs provided as type_id kwargs will be ignored for the
+        # criteria and will remove all the content in the repo.
+        # Refer to https://bugzilla.redhat.com/show_bug.cgi?id=1021579 and Pulp
+        # documentation for more details.
 
-        # start down the path of using Criteria per this issue:
-        # https://github.com/release-engineering/pubtools-pulplib/issues/62
-        # by using Criteria internally.
-        # there seems to be pretty complex handling of Criteria
-        # for serialization in the search API, and it is unclear which parts
-        # might also be necessary to use for content removal.
-        # If any or all of the same handling is needed, it would be beneficial
-        # to encapsulate the preparation of a criteria JSON object in some
-        # (more generically named) functions or a class to avoid duplicating code.
-        # for reference see search_content, _impl.client.Client._search_repo_units,
-        # _impl.client.Client._search, and _impl.client.search.search_for_criteria
-        criteria = None
-        type_ids = kwargs.get("type_ids")
-        # use _content_type_id field name to coerce
-        # search_for_criteria to fill out the PulpSearch#type_ids field.
-        # passing a criteria with an empty type_ids list rather than
-        # None results in failing tests due to the implementation of
-        # FakeClient#_do_unassociate
-        if type_ids is not None:
-            criteria = Criteria.with_field(
-                "_content_type_id",
-                Matcher.in_(type_ids),  # Criteria.with_field_in is deprecated
-            )
+        # Note: type_ids is deprecated. Criteria.with_unit_type should be used to
+        # filter on type_ids. This is kept for backward compatibility and will be
+        # removed in future versions.
+        if not criteria:
+            type_ids = kwargs.get("type_ids")
+            # use content_type_id field name to coerce
+            # search_for_criteria to fill out the PulpSearch#type_ids field.
+            # passing a criteria with an empty type_ids list rather than
+            # None results in failing tests due to the implementation of
+            # FakeClient#_do_unassociate
+            if type_ids is not None:
+                warnings.warn(
+                    "type_ids is deprecated, use criteria instead", DeprecationWarning
+                )
+                criteria = Criteria.with_field(
+                    "content_type_id",
+                    Matcher.in_(type_ids),  # Criteria.with_field_in is deprecated
+                )
 
         return f_proxy(self._client._do_unassociate(self.id, criteria=criteria))
 
