@@ -4,7 +4,11 @@ from .base import Unit, unit_type
 
 from ..attr import pulp_attrib
 from ... import compat_attr as attr
-from ..convert import frozenlist_or_none_sorted_converter, tolerant_timestamp
+from ..convert import (
+    frozenlist_or_none_sorted_converter,
+    tolerant_timestamp,
+    timestamp_converter,
+)
 from ..validate import optional_str, instance_of
 
 
@@ -125,6 +129,7 @@ class FileUnit(Unit):
         pulp_field="pulp_user_metadata.cdn_published",
         default=None,
         converter=tolerant_timestamp,
+        py_pulp_converter=timestamp_converter,
         validator=instance_of((datetime.datetime, type(None))),
     )
     """Approximate :class:`~datetime.datetime` in UTC at which this file first
